import os
import logging
from network_automation import environment
from mydict import MyDict
from netmiko import ConnectHandler, NetMikoAuthenticationException, NetMikoTimeoutException


class CiscoSSHDevice(object):
    """
    This class defines methods for fetching data from a Cisco device using NetMiko
    """
    def __init__(self, hostname, username=None, password=None, device_type='cisco_ios'):
        if not hostname:
            raise ValueError("hostname is mandatory")

        self.hostname = hostname

        # Username and passwords can be provided as parameters or as environment variables
        self.username = username or environment.get_cisco_username()
        self.password = password or environment.get_cisco_password()

        # Only print console output if the VERBOSE environment variable is set to True or 1
        self.verbose = os.getenv("VERBOSE", "0").lower() in ["1", "true"]

        netmiko_device = {
            'device_type': device_type,
            'ip': self.hostname,
            'username': self.username,
            'password': self.password,
            'secret': self.password
        }
        try:
            self.conn = ConnectHandler(**netmiko_device)
            msg = f"Successfully connected to {hostname}"
            logging.info(msg)
            if self.verbose:
                print(msg)
        except (NetMikoTimeoutException, NetMikoAuthenticationException) as e:
            raise ConnectionError(f"Failed to connect to {hostname}: {e}")

    def execute_show_command(self, command, parse=True, timeout=10):
        """
        This method executes a command on Cisco CLI and returns the result
        :param command: The command to run
        :param parse: Parse the output with textfsm (True)
        :param timeout: Set the timeout for executing the command and getting the result
        :return:
        """
        logging.info(f"Executing command '{command}' on {self.hostname}")

        result = self.conn.send_command(command, use_textfsm=parse, read_timeout=timeout)

        if type(result) is not str and self.verbose:
            print(f"Could not parse command result")

        return result

    def get_interface_details(self, timeout=30):
        """
        This method executes the 'show interface' command and returns the result parsed with textfsm
        :param timeout: Set the timeout for executing the command and getting the result
        :return:
        """
        interfaces = self.execute_show_command('show interface', timeout=timeout)

        return [MyDict(x) for x in interfaces]

    def get_cdp_neighbors(self, detail=False):
        """
        This method retrieves the CDP neighbors of the device
        :return:
        """
        if detail:
            command = 'show cdp neighbors detail'
        else:
            command = 'show cdp neighbors'

        return self.execute_show_command(command)

    def get_device_serial(self):
        """
        This method gets the serial number of a device
        :return:
        """
        serial = self.conn.send_command('show version | include Processor')

        return serial.split(' ')[-1]

    def get_ip_interface_brief(self):
        """
        This method gets interface IP information
        :return:
        """
        return [MyDict(x) for x in self.execute_show_command('show ip interface brief')]



import logging
import os
import pathlib
import sys
from dotenv import load_dotenv

# Get environment variables from local .env file and user's home directory .env file
dotenv_current_path = os.path.join(pathlib.Path().resolve(), '.env')
dotenv_home_path = os.path.join(pathlib.Path.home().resolve(), '.env')
load_dotenv(dotenv_home_path)
load_dotenv(dotenv_current_path)

# Logging config
if sys.version_info.major >= 3 and sys.version_info.minor >= 9:
    logging.basicConfig(filename='network_automation.log',
                        encoding='utf-8',
                        level=logging.INFO,
                        format='%(levelname)s:%(asctime)s %(message)s',
                        datefmt='%d/%m/%Y %H:%M:%S')
else:
    logging.basicConfig(filename='network_automation.log',
                        level=logging.INFO,
                        format='%(levelname)s:%(asctime)s %(message)s',
                        datefmt='%d/%m/%Y %H:%M:%S')

# GENERAL SETTINGS
VERBOSE = False
if 'VERBOSE' in os.environ and (
        os.environ["VERBOSE"].lower() == "true"
        or os.environ["VERBOSE"].lower() == "yes"
        or os.environ["VERBOSE"] == 1):
    VERBOSE = True

DEBUG = False
if 'DEBUG' in os.environ and (
        os.environ["DEBUG"].lower() == "true"
        or os.environ["DEBUG"].lower() == "yes"
        or os.environ["DEBUG"] == 1):
    DEBUG = True


def get_cisco_username():
    """Get username for Cisco devices from environment variables"""
    return os.environ.get("CISCO_USERNAME")


def get_cisco_password():
    """Get password for Cisco devices from environment variables"""
    return os.environ.get("CISCO_PASSWORD")


def get_netbox_url():
    """Get URL for NetBox from environment variables"""
    return os.environ.get("NETBOX_URL")


def get_netbox_token():
    """Get token for NetBox from environment variables"""
    return os.environ.get("NETBOX_TOKEN")



import json
import os
import requests
from mydict import MyDict


class Authentication:
    def __init__(self, host, port, usr, pwd, proxies):
        self.host = host or os.environ.get("VMANAGE_HOST")
        self.port = port or os.environ.get("VMANAGE_PORT", 443)
        self.usr = usr or os.environ.get("VMANAGE_USER")
        self.pwd = pwd or os.environ.get("VMANAGE_PASS")
        self.proxies = proxies or {}

        self.jsessionid = self.get_jsessionid()
        self.token = self.get_token()

    def get_jsessionid(self):
        api = "/j_security_check"
        base_url = f'https://{self.host}:{self.port}'
        url = base_url + api
        payload = {'j_username': self.usr, 'j_password': self.pwd}

        response = requests.post(url=url, data=payload, proxies=self.proxies, verify=False)
        try:
            cookies = response.headers["Set-Cookie"]
            jsessionid = cookies.split(";")
        except:
            raise ConnectionError("No valid JSESSION ID returned")

        return jsessionid[0]

    def get_token(self):
        headers = {'Cookie': self.jsessionid}
        base_url = f'https://{self.host}:{self.port}'
        api = "/dataservice/client/token"
        url = base_url + api
        response = requests.get(url=url, headers=headers, proxies=self.proxies, verify=False)
        if response.status_code != 200:
            raise ConnectionError("No valid token returned")

        return response.text


class VManage:
    def __init__(self, host=None, port=None, usr=None, pwd=None, proxies=None):
        self.auth = Authentication(host, port, usr, pwd, proxies)
        # self.jsessionid = self.auth.get_jsessionid(vmanage_host, vmanage_port, vmanage_username, vmanage_password)
        # self.token = self.auth.get_token(vmanage_host, vmanage_port, self.jsessionid)
        self.base_url = f'https://{self.auth.host}:{self.auth.port}/dataservice'
        self.proxies = self.auth.proxies

        if self.auth.token is not None:
            self.headers = {'Content-Type': "application/json", 'Cookie': self.auth.jsessionid,
                            'X-XSRF-TOKEN': self.auth.token}
        else:
            self.headers = {'Content-Type': "application/json", 'Cookie': self.auth.jsessionid}

    def get_all_devices(self):
        url_path = '/device'

        result = requests.get(self.base_url + url_path, headers=self.headers,
                              proxies=self.proxies, verify=False).json()

        return [MyDict(x) for x in result['data']]

    def get_prefix_lists(self):
        url_path = '/template/policy/list/dataprefix'

        result = requests.get(self.base_url + url_path, headers=self.headers,
                              proxies=self.proxies, verify=False).json()

        return [MyDict(x) for x in result['data']]

    def get_security_policies(self):
        url_path = '/template/policy/security'

        result = requests.get(self.base_url + url_path, headers=self.headers,
                              proxies=self.proxies, verify=False).json()

        return [MyDict(x) for x in result['data']]

    def get_zbf_policies(self):
        url_path = '/template/policy/definition/zonebasedfw'

        result = requests.get(self.base_url + url_path, headers=self.headers,
                              proxies=self.proxies, verify=False).json()

        return [MyDict(x) for x in result['data']]

    def get_zbf_policy(self, policy_id):
        url_path = f'/template/policy/definition/zonebasedfw/{policy_id}'

        return MyDict(requests.get(self.base_url + url_path, headers=self.headers,
                                   proxies=self.proxies, verify=False).json())

    def get_tunnel_metrics(self, device_ip, remote_endpoints, hours=24, interval=1):
        url_path = '/statistics/approute/fec/aggregation'
        query = {
            "query": {
                "condition": "AND",
                "rules":
                    [
                        {
                            "value": [str(hours)],
                            "field": "entry_time",
                            "type": "date",
                            "operator": "last_n_hours"
                        },
                        {
                            "value": [device_ip],
                            "field": "vdevice_name",
                            "type": "string",
                            "operator": "in"
                        }
                    ]
            },
            "aggregation":
                {
                    "field": [
                        {
                            "property": "name",
                            "sequence": 1,
                            "size": 262},
                        {
                            "property": "state",
                            "sequence": 1
                        },
                        {
                            "property": "proto",
                            "sequence": 2
                        }
                    ],
                    "histogram": {
                        "property": "entry_time",
                        "type": "hour",
                        "interval": interval,
                        "order": "asc"
                    },
                    "metrics": [
                        {
                            "property": "loss_percentage",
                            "type": "avg"},
                        {
                            "property": "vqoe_score",
                            "type": "avg"
                        },
                        {
                            "property": "latency",
                            "type": "avg"
                        },
                        {
                            "property": "jitter",
                            "type": "avg"
                        },
                        {
                            "property": "rx_octets",
                            "type": "sum"
                        },
                        {
                            "property": "tx_octets",
                            "type": "sum"}
                    ]
                }
        }

        result = MyDict(requests.post(self.base_url + url_path, headers=self.headers,
                                      proxies=self.proxies, data=json.dumps(query), verify=False).json())

        filtered_list = []

        for endpoint in remote_endpoints:
            filtered_list.extend([x for x in result.data if endpoint in x.name])

        return filtered_list



def ip_reachable(host):
    """
    Checks if a host responds to a ping request.
    :param host: Hostname or IP address to be checked
    :return: True if the host is alive, false otherwise
    """
    import platform
    import subprocess

    # Option for the number of packets as a function of
    count = '-n' if platform.system().lower() == 'windows' else '-c'
    wait = '-w' if platform.system().lower() == 'windows' else '-W'

    # Building the command. Ex: "ping -c 1 google.com"
    command = ['ping', count, '2', wait, '5', host]

    return subprocess.call(command, stdout=subprocess.DEVNULL) == 0



import os
import re
import requests
from network_automation import environment


class CiscoACI:
    def __init__(self, url, username=None, password=None):
        apic_auth_data = {
            "aaaUser": {
                "attributes": {
                    "name": username or environment.get_cisco_username(),
                    "pwd": password or environment.get_cisco_password()
                }
            }
        }

        self.url = url
        self.auth_url = self.url + "aaaLogin.json"

        self.session = requests.session()
        self.session.verify = False
        self.session.post(self.auth_url, json=apic_auth_data, verify=False)

    def get_aci_pods(self):
        fabric_class = "node/class/fabricPod.json"
        fabric_url = self.url + fabric_class

        aci_fabrics = self.session.get(fabric_url)
        return [x['fabricPod']['attributes']['dn'] for x in aci_fabrics.json()['imdata']]

    def get_aci_nodes(self, pod):
        node_class = f"node/mo/{pod}.json?query-target=children&target-subtree-class=fabricNode"
        node_url = self.url + node_class

        return [x['fabricNode']['attributes'] for x in self.session.get(node_url).json()['imdata']]

    def get_node_mgmt_ip(self, path):
        mgmt_intf_class = f"node/mo/{path}/sys/ipv4/inst/dom-management/if-[mgmt0].json?query-target=children&target-subtree-class=ipv4Addr"
        mgmt_intf_url = self.url + mgmt_intf_class

        result = self.session.get(mgmt_intf_url).json()

        try:
            mgmt_ip = result['imdata'][0]['ipv4Addr']['attributes']['addr']
        except (IndexError, KeyError):
            print(f"Could not get management IP address for {path}: {result}")
            return None

        return mgmt_ip

    def get_physical_intfs(self, path):
        physical_intfs_class = f"node/class/{path}/l1PhysIf.json?rsp-subtree=children&rsp-subtree-class=ethpmPhysIf"
        physical_intfs_url = self.url + physical_intfs_class
        physical_intfs = self.session.get(physical_intfs_url).json()

        result = [{
            'name': x['l1PhysIf']['attributes']['id'],
            'admin_status': x['l1PhysIf']['attributes']['adminSt'],
            'desc': x['l1PhysIf']['attributes']['descr'],
            'mode': x['l1PhysIf']['attributes']['mode'],
            'mac': x['l1PhysIf']['children'][0]['ethpmPhysIf']['attributes']['backplaneMac']
        } for x in physical_intfs['imdata']]

        return result

    def get_l3_loopbacks(self, path):
        loopback_intfs_class = f"node/class/{path}/l3LbRtdIf.json?query-target=children&target-subtree-class=ethpmLbRtdIf"
        loopback_intfs_url = self.url + loopback_intfs_class
        loopback_intfs = self.session.get(loopback_intfs_url).json()['imdata']

        for intf in loopback_intfs:
            intf_id_match = re.search(r'\[(.*?)\]', intf['ethpmLbRtdIf']['attributes']['dn'])
            if intf_id_match:
                intf['id'] = intf_id_match.group(1)
                loopback_ip_class = f"node/mo/{path}/sys/ipv4/inst/dom-overlay-1/if-[{intf['id']}].json?query-target=children&target-subtree-class=ipv4Addr"
                loopback_ip_url = self.url + loopback_ip_class
                try:
                    intf['ipv4_addr'] = self.session.get(loopback_ip_url).json()['imdata'][0]['ipv4Addr']['attributes']['addr']
                except KeyError:
                    print(f"Could not get IP address for interface {intf['id']}")

        return loopback_intfs

    def get_tenants(self):
        tenants = self._get_class("node/class/fvTenant.json")

        return [x['fvTenant']['attributes']['name'] for x in tenants]

    def get_tenant_bridge_domains(self, tenant_name):
        bds = self._get_class(f"node/mo/uni/{tenant_name}.json?query-target=children&target-subtree-class=fvBD")

        return [x['fvBD']['attributes']['name'] for x in bds]

    def _get_class(self, url_path):
        class_url = self.url + url_path

        return self.session.get(class_url).json()['imdata']



import json
import requests
from network_automation import environment

header = {
    "content-type": "application/vnd.yang.collection+json",
    "Accept": "application/vnd.yang.data+json",
}


class NFVISServer(object):
    """
    This represents a Cisco NFVIS server
    """
    def __init__(self, hostname, username=None, password=None, verify=True):
        if not hostname:
            raise ValueError("Hostname is missing")

        self.hostname = hostname
        self.url = 'https://' + hostname
        self.username = username or environment.get_cisco_username()
        self.password = password or environment.get_cisco_password()

        if not self.username or not self.password:
            raise ValueError("username/password is missing and could not be retrieved from environment variables")

        self.session = requests.Session()
        if not verify:
            self.session.verify = False

        self.session.auth = (self.username, self.password)

        self.platform = None
        self.get_platform_details()

    def get_platform_details(self):
        if self.platform:
            return self.platform

        uri = self.url + '/api/operational/platform-detail'
        try:
            resp = self.session.get(uri, headers=header)
        except requests.exceptions.ConnectionError:
            raise ConnectionError(f"Could not connect to {self.hostname}")

        self.platform = json.loads(resp.text)["platform_info:platform-detail"]

    def get_serial(self):
        try:
            return self.platform["hardware_info"]["SN"]
        except KeyError:
            return ""

    def get_pid(self):
        try:
            return self.platform["hardware_info"]["PID"]
        except KeyError:
            return ""

    def get_version(self):
        try:
            return self.platform["hardware_info"]["Version"]
        except KeyError:
            return ""

    def get_interfaces(self, detailed=False):
        if detailed:
            uri = self.url + '/api/operational/pnics?deep'
        else:
            uri = self.url + '/api/operational/pnics'

        try:
            resp = self.session.get(uri, headers=header)
        except requests.exceptions.ConnectionError:
            print("Could not get interface information")
            return None

        return json.loads(resp.text)["pnic:pnics"]["pnic"]

    def get_switch_interfaces(self, detailed=False):
        if detailed:
            uri = self.url + '/api/operational/switch/interface/status?deep'
        else:
            uri = self.url + '/api/operational/switch/interface/status'

        try:
            resp = self.session.get(uri, headers=header)
        except requests.exceptions.ConnectionError:
            print("Could not get switch interface information")
            return None

        if resp.status_code == 200:
            return json.loads(resp.text)["switch:status"]

        return []

    def get_switchport_status(self, detailed=False):
        if detailed:
            uri = self.url + '/api/operational/switch/interface/switchPort?deep'
        else:
            uri = self.url + '/api/operational/switch/interface/switchPort'

        try:
            resp = self.session.get(uri, headers=header)
        except requests.exceptions.ConnectionError:
            print("Could not get switchport information")
            return None

        if resp.status_code == 200:
            return json.loads(resp.text)["switch:switchPort"]

        return []



import ipaddress
import logging
from network_automation import environment
from collections import defaultdict
from pynetbox import api as netbox_api


class NetBoxInstance(netbox_api):
    """
    This class extends the pynetbox api class by adding additional methods that are not strictly related to NetBox
    data itself, but more like custom methods that help a manage data within NetBox
    """
    def __init__(self, url=None, token=None):
        self.url = url or environment.get_netbox_url()
        self.token = token or environment.get_netbox_token()

        super(NetBoxInstance, self).__init__(url=self.url, token=self.token)

    def duplicated_device_serials(self):
        """
        Check if there are multiple devices with the same serial. This should not happen normally
        :return: The duplicated serial numbers
        """
        duplicates = []
        seen_values = defaultdict(list)

        for entry in [x for x in self.dcim.devices.all() if x.serial]:
            serial = entry['serial']
            if seen_values[serial]:
                duplicates.extend(seen_values[serial])
            else:
                seen_values[serial].append(entry['serial'])

        return duplicates

    def get_ip_addresses_without_prefix(self):
        """
        This function returns all IP addresses that have no associated prefix
        :return:
        """
        # Get prefixes and IP addresses from NetBox
        prefixes = [x.prefix for x in self.ipam.prefixes.all()]
        ip_addresses = [x.address for x in self.ipam.ip_addresses.all()]

        result = []
        # Loop through the IP addresses and check each one
        for ip_address in ip_addresses:
            # Get the corresponding network
            subnet = ipaddress.ip_network(ip_address, False)
            if str(subnet) not in prefixes:
                logging.info(f"Adding {ip_address} to the list")
                result.append(str(subnet))

        return result

    def assign_primary_ip(self):
        """
        For devices that have only 1 IP address, that IP address will be assigned as primary IP
        :return:
        """
        # Get a list of all devices which don't have a primary IP
        devices_without_ip = [x for x in self.dcim.devices.all() if not x.primary_ip]

        for device in devices_without_ip:
            # Get a list of IP addresses assigned to the device
            device_ip_addresses = [x for x in self.ipam.ip_addresses.filter(device_id=device.id)]

            # If the list has one member, set that IP address to be the primary IP of the device
            if len(device_ip_addresses) == 1:
                only_ip = device_ip_addresses[0]
                print(f"Updating primary IP of {device} to {only_ip}...")
                if device.update({"primary_ip4": only_ip.id}):
                    logging.info(f"Updated primary IP of {device} to {only_ip}")
                else:
                    logging.error(f"Failed to update primary IP of {device}")
